
//////////////////////////////////////////////////////////////////////
//
// CyoRand.h
//
// Developed by Graham M. Bull for Cyotec Systems Limited.
// http://www.cyotec.com
//
// Copyright (c) 2003 by Cyotec Systems Limited. All Rights Reserved.
//
// This source code may be used, modified, compiled, duplicated, and/or
// distributed without restriction provided this copyright notice remains intact.
// Cyotec Systems Limited and/or its employees cannot be held responsible for any
// direct or indirect damage or loss of any kind that may result from using this
// code, and provides no warranty, guarantee, or support.
//
// History
// - June 19, 2003 - Created
//
//////////////////////////////////////////////////////////////////////

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <crtdbg.h>
#include <math.h>
#include <tchar.h>

#include <iostream>
#include <list>
using namespace std;

#include <shlwapi.h>
#pragma comment( lib, "shlwapi.lib" )

#define DIRECTSOUND_VERSION 0x0800
#include <mmsystem.h>
#include <dsound.h>
#pragma comment( lib, "dsound.lib" )

//////////////////////////////////////////////////////////////////////
// Types

typedef list< LPTSTR > StringList;
typedef list< LPTSTR >::iterator StringListIterator;
typedef unsigned __int64 QWORD;

//////////////////////////////////////////////////////////////////////
// Constants

const int s_nChannels = 2;
const int c_nSampleRate = 44100;
const WORD c_wBitsPerSample = 16;
const DWORD c_dwCapturePause = 100;

//////////////////////////////////////////////////////////////////////
// Macros

#define RELEASE(a) if (a != NULL) { a->Release(); a = NULL; }
#define FCLOSE(a) if (a != NULL) { fclose( a ); a = NULL; }

//////////////////////////////////////////////////////////////////////
// Prototypes

void Process1( LPBYTE pDest, LPWORD pSrc, DWORD dwSize );
void Process2( LPBYTE pDest, LPBYTE pSrc, DWORD dwSize );
void Process3( LPBYTE pDest, LPBYTE pSrc, DWORD dwSize );
LPCTSTR Append( LPCTSTR pszString1, LPCTSTR pszString2 );
void ClearFiles( StringList& rList );
void AnalyzeBuffer( LPBYTE pBuffer, DWORD dwSize, bool bListValues );
LPCTSTR FormatHugeNum( QWORD Num );

//////////////////////////////////////////////////////////////////////
// Functions

int _tmain( int argc, _TCHAR* argv[] )
{
	HRESULT hres = S_OK;
	LPDIRECTSOUNDCAPTURE8 pdsc = NULL;
	LPDIRECTSOUNDCAPTUREBUFFER pdscBuffer = NULL;
	TCHAR szPathname[ MAX_PATH ];
	FILE* pRawFile = NULL;
	FILE* pInputFile = NULL;
	FILE* pMidFile = NULL;
	FILE* pOutputFile = NULL;
	StringList m_listInputFiles;
	StringList m_listMidFiles;
	StringListIterator it;

	try
	{
		if ((argc != 5) && (argc != 6))
			throw _T("Usage:\n\n\tCYORAND path inputFiles midFiles seconds [/LIST]");

		LPTSTR pszPath = argv[ 1 ];
		if (!::PathFileExists( pszPath ))
			throw Append( _T("Invalid path - "), pszPath );

		int nInputFilesPerMidFile = _ttoi( argv[ 2 ]);
		if (nInputFilesPerMidFile < 1)
			throw _T("InputFiles must be >= 1");

		int nMidFilesPerOutputFile = _ttoi( argv[ 3 ]);
		if (nMidFilesPerOutputFile < 1)
			throw _T("MidFiles must be >= 1");

		int nSeconds = _ttoi( argv[ 4 ]);
		if (nSeconds < 1)
			throw _T("Seconds must be >= 1");

		bool bListValues = false;
		if (argc == 6)
		{
			if (_tcsicmp( argv[ 5 ], _T("/LIST") ) == 0)
				bListValues = true;
			else
				cerr << _T("Invalid argument - ") << argv[ 5 ] << endl;
		}

		//
		// Initialise...
		//
		hres = DirectSoundCaptureCreate8( NULL, &pdsc, NULL );
		if (FAILED( hres ))
			throw _T("Unable to create DirectSoundCapture8 interface");
		_ASSERT( pdsc != NULL );

		WAVEFORMATEX wfx;
		memset( &wfx, 0, sizeof( wfx ));
		wfx.wFormatTag = WAVE_FORMAT_PCM;
		wfx.nChannels = s_nChannels;
		wfx.nSamplesPerSec = c_nSampleRate;
		wfx.wBitsPerSample = c_wBitsPerSample;
		wfx.nBlockAlign = (wfx.nChannels * wfx.wBitsPerSample) / 8;
		wfx.nAvgBytesPerSec = (wfx.nSamplesPerSec * wfx.nBlockAlign);

		DSCBUFFERDESC dscBufferDesc;
		memset( &dscBufferDesc, 0, sizeof( dscBufferDesc ));
		dscBufferDesc.dwSize = sizeof( DSCBUFFERDESC );
		dscBufferDesc.dwBufferBytes = (wfx.nAvgBytesPerSec * nSeconds);
		dscBufferDesc.lpwfxFormat = &wfx;

		// Create sound buffer...
		hres = pdsc->CreateCaptureBuffer( &dscBufferDesc, &pdscBuffer, NULL );
		if (FAILED( hres ))
			throw _T("Unable to create buffer");
		_ASSERT( pdscBuffer != NULL );

		// Info...
		int nTotalInputFiles = (nInputFilesPerMidFile * nMidFilesPerOutputFile);
		DWORD dwRawSize = (dscBufferDesc.dwBufferBytes / 16);
		int nEstimatedTime = (int)((double)(nTotalInputFiles * nSeconds) / 60.0);
		cout << _T("\tFiles to record = ") << nTotalInputFiles << endl;
		cout << _T("\tEstimated time = ") << nEstimatedTime << _T(" minute(s)") << endl;
		cout << _T("\tOutput file size = ") << FormatHugeNum( dwRawSize ) << _T(" bytes") << endl;

		//
		// Capture...
		//
		cout << _T("Capturing...") << endl;
		for (int i = 0; i < nTotalInputFiles; ++i)
		{
			cout << _T('\t') << (nTotalInputFiles - i) << _T(" \r") << flush;

			hres = pdscBuffer->Start( 0 );
			if (FAILED( hres ))
				throw _T("Capture failed");
			for (;;)
			{
				DWORD dwStatus = 0;
				hres = pdscBuffer->GetStatus( &dwStatus );
				if (FAILED( hres ))
					throw _T("GetStatus failed");
				if ((dwStatus & DSCBSTATUS_CAPTURING) == 0)
					break;
				::Sleep( c_dwCapturePause );
			}

			// Lock sound buffer...
			LPVOID pBlock1, pBlock2;
			DWORD dwBlock1, dwBlock2;
			hres = pdscBuffer->Lock( 0, 0, &pBlock1, &dwBlock1, &pBlock2, &dwBlock2, DSCBLOCK_ENTIREBUFFER );
			if (FAILED( hres ))
				throw _T("Unable to lock buffer");

			// Process raw data...
			LPBYTE pRawData = new BYTE[ dwRawSize ];
			memset( pRawData, 0, dwRawSize );
			_ASSERT( (pBlock1 != NULL) && (pBlock2 == NULL) );
			Process1( pRawData, (LPWORD)pBlock1, dwBlock1 );

			// Save data...
			if (GetTempFileName( pszPath, _T("cap"), 0, szPathname ) == 0)
				throw _T("GetTempFileName failed");
			pRawFile = _tfopen( szPathname, _T("wb") );
			if (pRawFile == NULL)
				throw Append( _T("Unable to create raw file - "), szPathname );
			if (fwrite( pRawData, 1, dwRawSize, pRawFile ) != dwRawSize)
				throw Append( _T("Unable to write raw block - "), szPathname );
			FCLOSE( pRawFile );
			delete[] pRawData;
			m_listInputFiles.push_back( _tcsdup( szPathname ));

			// Unlock sound buffer...
			hres = pdscBuffer->Unlock( pBlock1, dwBlock1, pBlock2, dwBlock2 );
			if (FAILED( hres ))
				throw _T("Unable to unlock buffer");
		}

		//
		// Merge input files...
		//
		cout << _T("Merging input files...") << endl;
		LPBYTE pInputBuffer = new BYTE[ dwRawSize ];
		_ASSERT( pInputBuffer != NULL );
		LPBYTE pMidBuffer = new BYTE[ dwRawSize ];
		_ASSERT( pMidBuffer != NULL );
		memset( pMidBuffer, 0, dwRawSize );
		int nInputFile = 0;
		int nMidFile = 0;
		for (it = m_listInputFiles.begin(); it != m_listInputFiles.end(); ++it)
		{
			// Open and read input file
			pInputFile = _tfopen( *it, _T("rb") );
			if (pInputFile == NULL)
				throw Append( _T("Unable to open input file"), *it );
			if (fread( pInputBuffer, 1, dwRawSize, pInputFile ) != dwRawSize)
				throw Append( _T("Unable to read input file"), *it );
			FCLOSE( pInputFile );

			// Merge
			Process2( pMidBuffer, pInputBuffer, dwRawSize );

			++nInputFile;
			if (nInputFile >= nInputFilesPerMidFile)
			{
				// Open and write mid file
				if (GetTempFileName( pszPath, _T("mid"), 0, szPathname ) == 0)
					throw _T("GetTempFileName failed");
				pMidFile = _tfopen( szPathname, _T("wb") );
				if (pMidFile == NULL)
					throw Append( _T("Unable to open mid file - "), szPathname );
				if (fwrite( pMidBuffer, 1, dwRawSize, pMidFile ) != dwRawSize)
					throw Append( _T("Unable to write mid file - "), szPathname );
				FCLOSE( pMidFile );
				m_listMidFiles.push_back( _tcsdup( szPathname ));

				nInputFile = 0;
				++nMidFile;
			}
		}
		delete[] pInputBuffer;

		//
		// Merge mid files...
		//
		cout << _T("Merging mid files...") << endl;
		LPBYTE pOutputBuffer = new BYTE[ dwRawSize ];
		_ASSERT( pOutputBuffer != NULL );
		memset( pOutputBuffer, 0, dwRawSize );
		for (it = m_listMidFiles.begin(); it != m_listMidFiles.end(); ++it)
		{
			// Open and read mid file
			pMidFile = _tfopen( *it, _T("rb") );
			if (pMidFile == NULL)
				throw Append( _T("Unable to open mid file - "), *it );
			if (fread( pMidBuffer, 1, dwRawSize, pMidFile ) != dwRawSize)
				throw Append( _T("Unable to read mid file - "), *it );
			FCLOSE( pMidFile );

			// Merge
			Process3( pOutputBuffer, pMidBuffer, dwRawSize );
		}
		delete[] pMidBuffer;

		//
		// Write output file...
		//
		if (GetTempFileName( pszPath, _T("dat"), 0, szPathname ) == 0)
			throw _T("GetTempFileName failed");
		pOutputFile = _tfopen( szPathname, _T("wb") );
		if (pOutputFile == NULL)
			throw Append( _T("Unable to open output file - "), szPathname );
		if (fwrite( pOutputBuffer, 1, dwRawSize, pOutputFile ) != dwRawSize)
			throw Append( _T("Unable to write output file - "), szPathname );
		FCLOSE( pOutputFile );
		cout << _T('\t') << szPathname << endl;
		cout << _T("Analysis...") << endl;
		AnalyzeBuffer( pOutputBuffer, dwRawSize, bListValues );
		delete[] pOutputBuffer;
	}
	catch (LPCTSTR pszError)
	{
		cerr << pszError << endl;
	}

	// Close any open files...
	FCLOSE( pRawFile );
	FCLOSE( pInputFile );
	FCLOSE( pMidFile );
	FCLOSE( pOutputFile );

	// Delete temporary files...
	ClearFiles( m_listInputFiles );
	ClearFiles( m_listMidFiles );

	// Release DirectSoundCaptureBuffer...
	RELEASE( pdscBuffer );

	// Release DirectSoundCapture...
	RELEASE( pdsc );

	return 0;
}

// Raw captured data to InputFile
void Process1( LPBYTE pDest, LPWORD pSrc, DWORD dwSize )
{
	_ASSERT( pDest != NULL );
	_ASSERT( pSrc != NULL );

	int nShift = 0;
	for (LPWORD pSentinel = (LPWORD)((LPBYTE)pSrc + dwSize); pSrc < pSentinel; ++pSrc)
	{
		*pDest |= (((LOBYTE( *pSrc ) ^ HIBYTE( *pSrc )) & 1) << nShift);
		if (++nShift >= 8)
		{
			nShift = 0;
			++pDest;
		}
	}
}

// InputFile to MidFile
void Process2( LPBYTE pDest, LPBYTE pSrc, DWORD dwSize )
{
	_ASSERT( pDest != NULL );
	_ASSERT( pSrc != NULL );

	for (LPBYTE pSrcSentinel = (pSrc + dwSize); pSrc < pSrcSentinel; ++pSrc, ++pDest)
	{
		*pDest ^= *pSrc;
	}
}

// MidFile to OutputFile
void Process3( LPBYTE pDest, LPBYTE pSrc, DWORD dwSize )
{
	_ASSERT( pDest != NULL );
	_ASSERT( pSrc != NULL );

	for (LPBYTE pSrcSentinel = (pSrc + dwSize); pSrc < pSrcSentinel; ++pSrc, ++pDest)
	{
		static int nRotate = 0;
		*pDest ^= ((*pSrc << nRotate) | (*pSrc >> (8 - nRotate)));
		nRotate = (nRotate + 1) % 8;
	}
}

// Append two strings
LPCTSTR Append( LPCTSTR pszString1, LPCTSTR pszString2 )
{
	static TCHAR szOutput[ 512 ];
	_tcscpy( szOutput, pszString1 );
	_tcscat( szOutput, pszString2 );
	return szOutput;
}

// Remove all temporary files
void ClearFiles( StringList& rList )
{
	for (StringListIterator it = rList.begin(); it != rList.end(); ++it)
	{
		DeleteFile( *it );
		free( *it );
	}
}

// Display statistics for output data
void AnalyzeBuffer( LPBYTE pBuffer, DWORD dwSize, bool bListValues )
{
	const int c_nBytes = 256;

	long adwCount[ c_nBytes ];
	memset( adwCount, 0, c_nBytes * sizeof( long ));

	LPBYTE pSentinel = (pBuffer + dwSize);
	for (LPBYTE pCurr = pBuffer; pCurr < pSentinel; ++pCurr)
	{
		++adwCount[ *pCurr ];
	}

	// Sums, means
	long lNumBytes = 0, lSumOfBytes = 0;
	for (int i = 0; i < c_nBytes; ++i)
	{
		lNumBytes += adwCount[ i ];
		lSumOfBytes += (adwCount[ i ] * i);
	}
	double dMeanByte = ((double)lSumOfBytes / (double)lNumBytes);

	if (bListValues)
	{
		for (i = 0; i < c_nBytes; ++i)
		{
			TCHAR szData[ 128 ];
			_stprintf( szData, _T("%3i %-i %.3f"),
				i, adwCount[ i ], (double)adwCount[ i ] / (double)lNumBytes * 100.0 );

			TCHAR szOutput[ 128 ];
			_stprintf( szOutput, _T("%-20s"), szData );
			cout << szOutput << flush;
		}
	}

	cout << _T("\tMean byte = ") << dMeanByte << endl;
}

// Insert commas into large integers
LPCTSTR FormatHugeNum( QWORD qwNum )
{
	static TCHAR szNum[ 128 ];
	_ui64tot( qwNum, szNum, 10 );
	_tcsrev( szNum );
	int nCount = 0;
	for (TCHAR* pszCurr = szNum; *pszCurr != _T('\x0'); ++pszCurr)
	{
		nCount = (nCount + 1) % 4;
		if (nCount == 0)
		{
			TCHAR* pszNext = (pszCurr + 1);
			memmove( pszNext, pszCurr, _tcslen( pszCurr ) + 1 ); //+1 for terminator
			*pszCurr = _T(',');
		}
	}
	_tcsrev( szNum );
	return szNum;
}
